from typing import Any

from calculator.constants import SUPPORTED_INPUT_DATATYPES


class CalculatorException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidInputError(CalculatorException):
    def __init__(self, value: Any) -> None:
        message = (
            f"Invalid datatype error, {type(value).__name__}. "
            f"Supported datatypes {[input_datatype.__name__ for input_datatype in SUPPORTED_INPUT_DATATYPES]}."
        )
        super().__init__(message)


class NegativeValueRootError(CalculatorException):
    def __init__(self, value: Any) -> None:
        message = f"Root of a negative value is not supported, current value is {value}."
        super().__init__(message)


class DivisionByZeroError(CalculatorException):
    def __init__(self) -> None:
        message = f"Division by zero is not allowed."
        super().__init__(message)
