import pytest
from calculator.calculator import Calculator
from calculator.exceptions import (
    DivisionByZeroError,
    InvalidInputError,
    NegativeValueRootError,
)


@pytest.fixture(scope="function")
def calculator():
    return Calculator()


@pytest.fixture(scope="function")
def calculator_2(calculator):
    calculator.add(1)
    return calculator


class TestMultipleActions:
    def test_multiple_actions(self, calculator):
        # testing
        # ((0 + 1) / 1 - 1) * 1 = 0

        assert calculator.result == 0
        assert calculator.add(1) == 1
        assert calculator.divide(1) == 1
        assert calculator.subtract(1) == 0
        assert calculator.multiply(1) == 0
        assert calculator.result == 0


class TestResetMethod:
    def test_reset_after_add_method_call(self, calculator):
        calculator.add(1)
        calculator.reset()
        assert calculator.result == 0.0

    def test_reset_after_subtract_method_call(self, calculator):
        calculator.subtract(1)
        calculator.reset()
        assert calculator.result == 0.0

    def test_reset_after_multipy_method_call(self, calculator):
        calculator.multiply(1)
        calculator.reset()
        assert calculator.result == 0.0

    def test_reset_after_divide_method_call(self, calculator):
        calculator.divide(1)
        calculator.reset()
        assert calculator.result == 0.0


class TestGreenPathOfCalculatorMethods:
    def test_add(self, calculator):
        assert calculator.result == 0
        assert calculator.add(1) == 1
        assert calculator.add(0) == 1
        assert calculator.add(-1) == 0

    def test_subtract(self, calculator):
        assert calculator.result == 0
        assert calculator.subtract(1) == -1
        assert calculator.subtract(0) == -1
        assert calculator.subtract(-1) == 0

    def test_multiply(self, calculator_2):
        assert calculator_2.result == 1
        assert calculator_2.multiply(1) == 1
        assert calculator_2.multiply(-1) == -1
        assert calculator_2.multiply(0) == 0

    def test_divide(self, calculator_2):
        assert calculator_2.result == 1
        assert calculator_2.divide(1) == 1
        assert calculator_2.divide(-1) == -1


@pytest.mark.parametrize(
    "value",
    ["123", [99, 9], False, {99, 9}],
    ids=lambda param: f"input type: {type(param).__name__}",
)
class TestIncorrectInputs:
    match = "Invalid datatype error, {}. Supported datatypes ['int', 'float']."

    def test_add(self, calculator, value):
        # with pytest.raises(InvalidInputError, match=self.match.format(type(value).__name__)):
        with pytest.raises(InvalidInputError):
            calculator.add(value)

    def test_subtract(self, calculator, value):
        with pytest.raises(InvalidInputError):
            calculator.subtract(value)

    def test_multiply(self, calculator, value):
        with pytest.raises(InvalidInputError):
            calculator.multiply(value)

    def test_divide(self, calculator, value):
        with pytest.raises(InvalidInputError):
            calculator.divide(value)


class TestFunctionalErrors:
    def test_root_method_should_raise_NegativeValueRootError_when_current_value_is_negative(
        self, calculator
    ):
        with pytest.raises(
            NegativeValueRootError,
            match="Root of a negative value is not supported, current value is -1.",
        ):
            calculator.subtract(1)
            calculator.root(2)

    def test_divide_method_should_raise_NegativeValueRootError_when_current_value_is_negative(
        self, calculator
    ):
        with pytest.raises(
            DivisionByZeroError, match="Division by zero is not allowed."
        ):
            calculator.divide(0)
